/*
   Parallel search for "aardvark" in the files in the current directory.
   Each file is memory mapped and then searched in parallel.
*/

import java.time.Instant;
import java.time.Duration;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;
import java.util.concurrent.TimeUnit;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.Path;
import java.util.*;
import java.io.RandomAccessFile;
//import java.nio.MappedByteBuffer;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

public class SearchParallel_v8_MemoryMappedFiles
{
   public static int totalLines = 0; // There is a race condition on this variable!.

   public static void main(String[] args) throws IOException
   {
      final SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
      System.out.println(ft.format(new Date()) + " Search Parallel ver 8 in all Files");
      Instant start = Instant.now();

      Files.list(Paths.get("."))
             .filter(p -> p.toString().endsWith(".txt"))
               .forEach(p -> ForkJoinPool.commonPool().execute(new Task2(p))); // files processed in parallel

      Instant stop = Instant.now();
      long time = Duration.between(start, stop).toMillis();
      System.out.printf("%s All files submitted to the fork/join pool.\n", ft.format(new Date()));

      ForkJoinPool.commonPool().shutdown();
      try
      {
         ForkJoinPool.commonPool().awaitTermination(5, TimeUnit.MINUTES);
      }
      catch (InterruptedException e)
      {
         ForkJoinPool.commonPool().shutdownNow();
         Thread.currentThread().interrupt(); // Preserve interrupt status
      }

      stop = Instant.now();
      time = Duration.between(start, stop).toMillis();
      System.out.printf("%s processed %,d lines from all files.\n",
                       ft.format(new Date()), totalLines);
      System.out.printf("elapsed time: %,dms\n", time);
      System.out.println( ForkJoinPool.commonPool().toString() );
   }
}//main()


// A Task2 object, when it is called by one of the threads in
// the Fork/Join thread pool, will search one buffer of text.
@SuppressWarnings("serial")
class Task2 extends RecursiveAction
{
   private static final SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
   private static final int BUFFERSIZE = 16*8192;  // 8 * 8192;

   private Path path;

   public Task2(Path path)
   {
      this.path = path;
   }

   public void compute()
   {
      int lineNumber = 1;

      try
      {
         String fileName = path.getFileName().toString();
         FileChannel fc = new RandomAccessFile(fileName, "r").getChannel();
         long fileSize = fc.size(); // assume that each file is < 2GB

         for (long i = 0; i < fileSize; i += BUFFERSIZE)
         {
            // account for the last, partially full, buffer
            long bufferSize = (i + BUFFERSIZE < fileSize) ? BUFFERSIZE : fileSize - i;

            ByteBuffer mbb = fc.map( FileChannel.MapMode.READ_ONLY,
                                     i,
                                    (int)bufferSize );
            (new Task(mbb, lineNumber, fileName)).fork();
            for (int j = 0; j < mbb.capacity(); ++j)
            {
               if ( '\n' == mbb.get(j) )
               {
                  ++lineNumber; // count lines here
               }
            }
         }
         fc.close();
         SearchParallel_v8_MemoryMappedFiles.totalLines += lineNumber; // race condition on totalLines!
      }
      catch(IOException e){throw new Error(e);}
   }
}//Task2


// A Task object, when it is called by one of the threads in
// the Executor's thread pool, will filter one line of text.
@SuppressWarnings("serial")
class Task extends RecursiveAction
{
   private static final SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
   private final ByteBuffer buffer;
   private int lineNumber;
   private final String fileName;

   public Task(final ByteBuffer buffer, final int lineNumber, final String fileName)
   {
      this.buffer = buffer;
      this.lineNumber = lineNumber;
      this.fileName = fileName;
   }

   public void compute()
   {
      // Filter the data in the buffer.
      for (int i = 0; i < buffer.capacity(); ++i)
      {
         if ( '\n' == buffer.get(i) )
         {
            ++lineNumber;
         }
         else if ( i < buffer.capacity() - 8
                && 'a' == buffer.get(i)
                && 'a' == buffer.get(i+1)
                && 'r' == buffer.get(i+2)
                && 'd' == buffer.get(i+3)
                && 'v' == buffer.get(i+4)
                && 'a' == buffer.get(i+5)
                && 'r' == buffer.get(i+6)
                && 'k' == buffer.get(i+7) )
         {
            System.out.printf("%s line %,d in %s (current total line no %,d).\n",
                              ft.format(new Date()),
                              lineNumber,
                              fileName,
                              lineNumber + SearchParallel_v8_MemoryMappedFiles.totalLines);
         }
      }
   }
}//Task
